import asyncio
from contextlib import asynccontextmanager

import bleak

from . import driver
from . import protocol

async def discover(timeout=1):
    devices = await bleak.discover(
        timeout=timeout,
        filters={'UUIDs': [protocol.root_identifier_uuid]},
    )
    return [Robot(device) for device in devices]


@asynccontextmanager
async def get_robot(timeout=1):
    robots = await discover(timeout=timeout)
    async with robots[0] as robot:
        yield robot


async def run_robot(timeout=1, **callbacks):
    async with get_robot(timeout=timeout) as robot:
        robot.events.set_callbacks(**callbacks)
        await robot.events.process()


class Robot:
    def __init__(self, device):
        self._device = device
        self._client = None
        self._driver = None

        self.events = None
        self.motor = None
        self.marker = None
        self.eraser = None
        self.led = None
        self.music = None

    def _get_characteristics(self):
        uart = self._client.services.get_service(protocol.uart_service_uuid)
        rx = uart.get_characteristic(protocol.rx_char_uuid)
        assert 'write' in rx.properties
        tx = uart.get_characteristic(protocol.tx_char_uuid)
        assert 'notify' in tx.properties
        return rx, tx

    async def __aenter__(self):
        self._client = bleak.BleakClient(self._device)
        await self._client.__aenter__()

        rx, tx = self._get_characteristics()
        self._driver = driver.Driver(self._client, rx, tx)
        await self._driver.__aenter__()

        self.events = RobotEvents(self)
        self.motor = RobotMotor(self)
        self.marker = RobotMarker(self)
        self.eraser = RobotEraser(self)
        self.led = RobotLED(self)
        self.music = RobotMusic(self)

        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._driver.__aexit__(exc_type, exc, tb)
        self._driver = None
        await self._client.__aexit__(exc_type, exc, tb)
        self._client = None

    async def get_name(self):
        return await self._driver.get_name()

    async def set_name(self, name):
        await self._driver.set_name(name)

    async def get_main_board_version(self):
        return await self._driver.get_version(driver.Board.MAIN)

    async def get_color_board_version(self):
        return await self._driver.get_version(driver.Board.COLOR)

    async def get_serial_number(self):
        return await self._driver.get_serial_number()

    async def get_sku(self):
        return await self._driver.get_sku()

    async def get_battery_level(self):
        return await self._driver.get_battery_level()

    async def cancel(self):
        await self._driver.cancel()

    async def disconnect(self):
        await self._driver.disconnect()


class _RobotComponent:
    def __init__(self, robot):
        self._robot = robot
        self._driver = robot._driver


class RobotEvents(_RobotComponent):
    def __init__(self, robot):
        super().__init__(robot)
        self._callbacks = {}

    async def _iter_events(self, loop):
        async for event in self._driver.get_events(loop=loop):
            callback = self._callbacks.get(event.event_name, None)
            if callback is not None:
                asyncio.create_task(callback(self._robot, *event))

            yield event

    def set_callback(self, event_type, callback):
        if isinstance(event_type, driver._Event):
            event_name = event_type.event_name
        else:
            event_name = event_type

        self._callbacks[event_name] = callback

    def set_callbacks(self, **callbacks):
        self._callbacks.update(callbacks)

    async def __aiter__(self):
        "Iterate over all events, waiting for new ones"
        async for event in self._iter_events(True):
            yield event

    @property
    async def current(self):
        "Iterate over currently received events"
        async for event in self._iter_events(False):
            yield event

    async def process(self, wait=True, loop=True):
        """
        Process events (invoking callbacks)

        blocking if wait is true
        wait for new events if loop is true
        """
        if not wait:
            return asyncio.create_task(self.process(loop=loop))

        async for event in self._iter_events(loop):
            pass


class RobotMotor(_RobotComponent):
    async def set_speed(self, left, right):
        await self._driver.set_motor_speed(left, right)

    async def set_left_speed(self, speed):
        await self._driver.set_left_motor_speed(speed)

    async def set_right_speed(self, speed):
        await self._driver.set_right_motor_speed(speed)

    async def disable_gravity_compensation(self):
        await self._driver.set_gravity_compensation(driver.GravityState.OFF, 0)

    async def enable_gravity_compensation(self, amount=500):
        await self._driver.set_gravity_compensation(driver.GravityState.ON, amount)

    async def enable_gravity_compensation_on_marker(self, amount=500):
        await self._driver.set_gravity_compensation(driver.GravityState.ON_MARKER, amount)

    async def drive(self, distance, wait=True):
        await self._driver.drive_distance(distance, wait=True)

    async def rotate(self, angle, wait=True):
        await self._driver.rotate_angle(angle, wait=wait)

    async def drive_arc(self, angle: int, radius: int, wait=True):
        await self._driver.drive_arc(angle, radius, wait=wait)


class RobotMarker(_RobotComponent):
    async def down(self, wait=True):
        await self._driver.set_marker_eraser(driver.MarkerEraserPosition.MARKER_DOWN, wait=wait)

    async def up(self, wait=True):
        await self._driver.set_marker_eraser(driver.MarkerEraserPosition.UP, wait=wait)


class RobotEraser(_RobotComponent):
    async def down(self, wait=True):
        await self._driver.set_marker_eraser(driver.MarkerEraserPosition.ERASER_DOWN, wait=wait)

    async def up(self, wait=True):
        await self._driver.set_marker_eraser(driver.MarkerEraserPosition.UP, wait=wait)


class RobotLED(_RobotComponent):
    def __init__(self, robot):
        super().__init__(robot)
        self._anim = None
        self._color = (0, 0, 0)

    @property
    def color(self):
        return self._color

    async def _update(self, anim, color=None):
        self._anim = anim
        if color is not None:
            self._color = color

        r, g, b = self._color
        await self._driver.set_led_animation(self._anim, r, g, b)

    async def off(self):
        await self._update(driver.LEDAnimation.OFF)

    async def on(self, color=None):
        await self._update(driver.LEDAnimation.ON, color)

    async def blink(self, color=None):
        await self._update(driver.LEDAnimation.BLINK, color)

    async def spin(self, color=None):
        await self._update(driver.LEDAnimation.SPIN, color)


class RobotMusic(_RobotComponent):
    # + handle notes

    async def play(self, frequency, duration=1000):
        await self._driver.play_note(frequency, duration)

    async def stop(self):
        await self._driver.stop_note()

    async def say(self, phrase, wait=True):
        # + handle long phrase
        await self._driver.say_phrase(phrase, wait=wait)



import asyncio
import enum

from . import protocol


class Board(enum.Enum):
    MAIN = 0xA5
    COLOR = 0xC6


class GravityState(enum.Enum):
    OFF = 0
    ON = 1
    ON_MARKER = 2


class Motor(enum.Enum):
    LEFT = 0
    RIGHT = 1
    MARKER = 2


class MarkerEraserPosition(enum.Enum):
    UP = 0
    MARKER_DOWN = 1
    ERASER_DOWN = 2


class ColorSensor(enum.Enum):
    LEFT = 0
    CENTER_LEFT = 1
    CENTER_RIGHT = 2
    RIGHT = 3


class ColorLightning(enum.Enum):
    OFF = 0
    RED = 1
    GREEN = 2
    BLUE = 3
    ALL = 4


class ColorFormat(enum.Enum):
    ADC = 0
    MV = 1


class LEDAnimation(enum.Enum):
    OFF = 0
    ON = 1
    BLINK = 2
    SPIN = 3


class Bumper(enum.Flag):
    NO = 0
    RIGHT = 0x40
    LEFT = 0x80
    BOTH = 0xC0


class _Event(tuple):
    event_name = 'event'
    __fields__ = ()
    _register = {}

    def __new__(cls, *args):
        args = cls._parse_args(*args)
        return super().__new__(cls, args)

    def __init_subclass__(cls):
        import inspect
        import operator
        if cls.__fields__ is None:
            sig = inspect.signature(cls._parse_args)
            cls.__fields__ = tuple(sig.parameters)

        for i, field in enumerate(cls.__fields__):
            setattr(cls, field, property(operator.itemgetter(i)))

        cls._register[cls.event_name] = cls

    @classmethod
    def from_name(cls, event_name, *args):
        return cls._register[event_name](*args)

    @staticmethod
    def _parse_args():
        return ()

    def __repr__(self):
        args = ', '.join(f'{name}={value}' for name, value in zip(self.__fields__, self))
        return f'{self.__class__.__qualname__}({args})'


class StartedEvent(_Event):
    event_name = 'started'


class VersionReponse(_Event):
    event_name = 'version'
    __fields__ = None

    @staticmethod
    def _parse_args(board, fw_maj, fw_min, hw_maj, hw_min, boot_maj, boot_min, proto_maj, proto_min):
        return Board(board), fw_maj, fw_min, hw_maj, hw_min, boot_maj, boot_min, proto_maj, proto_min


class NameResponse(_Event):
    event_name = 'name'
    __fields__ = None

    @staticmethod
    def _parse_args(name):
        return name.decode(protocol.encoding),


class StoppedEvent(_Event):
    event_name = 'stopped'


class EnabledEventsResponse(_Event):
    event_name = 'enabled_events'
    __fields__ = None

    @staticmethod
    def _parse_args(data):
        return data,


class SerialNumberResponse(_Event):
    event_name = 'serial_number'
    __fields__ = None

    @staticmethod
    def _parse_args(serial_number):
        return serial_number.decode(protocol.encoding),


class SKUEvent(_Event):
    event_name = 'sku'
    __fields__ = None

    @staticmethod
    def _parse_args(sku):
        return sku.decode(protocol.encoding),


class DriveDistanceResponse(_Event):
    event_name = 'drive_distance_finished'


class RotateAngleResponse(_Event):
    event_name = 'rotate_angle_finished'


class DriveArcResponse(_Event):
    event_name = 'drive_arc_finished'


class MotorStallEvent(_Event):
    event_name = 'motor_stall'
    __fields__ = None

    @staticmethod
    def _parse_args(timestamp, motor, cause):
        return timestamp, Motor(motor), cause


class MarkerEraserResponse(_Event):
    event_name = 'marker_erase_finished'
    __fields__ = None

    @staticmethod
    def _parse_args(position):
        return MarkerEraserPosition(position),


class ColorResponse(_Event):
    event_name = 'color_response'
    __fields__ = None

    @staticmethod
    def _parse_args(a, b, c, d, e, f, g, h):
        return a, b, c, d, e, f, g, h


class ColorEvent(_Event):
    event_name = 'color_event'
    __fields__ = None

    @staticmethod
    def _parse_args(data):
        return data,


class PlayNoteResponse(_Event):
    event_name = 'play_note_finished'


class SayPhraseResponse(_Event):
    event_name = 'say_phrase_finished'


class BumperEvent(_Event):
    event_name = 'bumper_event'
    __fields__ = None

    @staticmethod
    def _parse_args(timestamp, state):
        return timestamp, Bumper(state)


class LightEvent(_Event):
    event_name = 'light_event'
    __fields__ = None

    @staticmethod
    def _parse_args(timestamp, state, left, right):
        return timestamp, state, left, right


class BatteryEvent(_Event):
    event_name = 'battery_event'
    __fields__ = None

    @staticmethod
    def _parse_args(timestamp, voltage, percent):
        return timestamp, voltage, percent


class BatteryResponse(_Event):
    event_name = 'battery_response'
    __fields__ = None

    @staticmethod
    def _parse_args(timestamp, voltage, percent):
        return timestamp, voltage, percent


class TouchEvent(_Event):
    event_name = 'touch_event'
    __fields__ = None

    @staticmethod
    def _parse_args(timestamp, state):
        return timestamp, state


class CliffEvent(_Event):
    event_name = 'cliff_event'
    __fields__ = None

    @staticmethod
    def _parse_args(timestamp, cliff, sensor, threshold):
        return timestamp, cliff, sensor, threshold


class Driver:
    def __init__(self, client, rx, tx):
        self.client = client
        self.rx = rx
        self.tx = tx
        self._responses = {}
        self._event_queue = asyncio.Queue()
        self._event_queue.put_nowait(StartedEvent())

    async def __aenter__(self):
        await self.client.start_notify(self.tx, self._notification)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.client.stop_notify(self.tx)

    async def _send(self, cmd, *args, wait_response=True):
        message, hdr = protocol.format_command(cmd, *args)
        if wait_response:
            self._responses[hdr] = waiter = asyncio.Event()
        await self.client.write_gatt_char(self.rx, message)
        if wait_response:
            await waiter.wait()
            return self._responses.pop(hdr)

    def _notification(self, _sender, message):
        name, args, hdr = protocol.extract_event(message)
        waiter = self._responses.pop(hdr, None)

        event = _Event.from_name(name, *args)
        if waiter is None:
            self._event_queue.put_nowait(event)
        else:
            self._responses[hdr] = event
            waiter.set()

    async def get_events(self, loop=False):
        # + handle disconnection
        q = self._event_queue
        while loop or not q.empty():
            event = await q.get()
            q.task_done()
            yield event

    async def get_version(self, board: Board):
        version = await self._send('get_version', board.value)
        return version

    async def set_name(self, name: str):
        await self._send('set_name', name.encode(protocol.encoding), wait_response=False)

    async def get_name(self):
        name, = await self._send('get_name')
        return name

    async def cancel(self):
        await self._send('cancel', wait_response=False)

    async def disconnect(self):
        await self._send('cancel', wait_response=False)

    async def enable_events(self, bitfield: bytes):
        await self._send('enable_events', bitfield, wait_response=False)

    async def disable_events(self, bitfield: bytes):
        await self._send('disable_events', bitfield, wait_response=False)

    async def get_enabled_events(self):
        bitfield, = await self._send('get_enabled_events')
        return bitfield

    async def get_serial_number(self):
        serial, = await self._send('get_serial_number')
        return serial

    async def get_sku(self):
        sku, = await self._send('get_sku')
        return sku

    async def set_motor_speed(self, left: int, right: int):
        await self._send('set_motor_speed', left, right, wait_response=False)

    async def set_left_motor_speed(self, speed: int):
        await self._send('set_left_motor_speed', speed, wait_response=False)

    async def set_right_motor_speed(self, speed: int):
        await self._send('set_right_motor_speed', speed, wait_response=False)

    async def set_gravity_compensation(self, state: GravityState, amount: int):
        await self._send('set_gravity_compensation', state.value, amount, wait_response=False)

    async def drive_distance(self, dist: int, wait=True):
        await self._send('drive_distance', dist, wait_response=wait)

    async def rotate_angle(self, angle: int, wait=True):
        await self._send('rotate_angle', angle, wait_response=wait)

    async def drive_arc(self, angle: int, radius: int, wait=True):
        await self._send('drive_arc', angle, radius, wait_response=wait)

    async def set_marker_eraser(self, position: MarkerEraserPosition, wait=True):
        ret = await self._send('set_marker_eraser', position.value, wait_response=wait)
        if wait:
            position, = ret
            return MarkerEraserPosition(position)

    async def get_color_data(self, sensor: ColorSensor, lightning: ColorLightning, format: ColorFormat):
        return await self._send('get_color_data', sensor.value, lightning.value, format.value)

    async def set_led_animation(self, anim: LEDAnimation, red: int, green: int, blue: int):
        await self._send('set_led_animation', anim.value, red, green, blue, wait_response=False)

    async def play_note(self, frequency: int, duration: int, wait=True):
        await self._send('play_note', frequency, duration, wait_response=wait)

    async def stop_note(self):
        await self._send('stop_note', wait_response=False)

    async def say_phrase(self, phrase: str, wait=True):
        await self._send('say_phrase', phrase.encode(protocol.encoding), wait_response=wait)

    async def get_battery_level(self):
        _, voltage, rate = await self._send('get_battery_level')
        return voltage, rate



import itertools
import struct

from crc8 import crc8


FULL_PACKET_LEN = 20
SIMPLE_PACKET_LEN = FULL_PACKET_LEN - 1

_msg_prefix = '>3B'
_msg_prefix_len = struct.calcsize(_msg_prefix)

root_identifier_uuid = '48c5d828-ac2a-442d-97a3-0c9822b04979'
uart_service_uuid = '6e400001-b5a3-f393-e0a9-e50e24dcca9e'
rx_char_uuid = '6e400002-b5a3-f393-e0a9-e50e24dcca9e'
tx_char_uuid = '6e400003-b5a3-f393-e0a9-e50e24dcca9e'
encoding = 'utf-8'

_commands = {
    # name: (dev, cmd, fmt)

    'get_version': (0, 0, 'B'),  # board (A5=main|C6=color)
    'set_name': (0, 1, '16s'),  # name
    'get_name': (0, 2, ''),
    'cancel': (0, 3, ''),
    'disconnect': (0, 6, ''),
    'enable_events': (0, 7, '16s'),  # devices bitfield (1=enable|0=unchange)
    'disable_events': (0, 9, '16s'),  # devices bitfield (1=disable|0=unchange)
    'get_enabled_events': (0, 11, ''),
    'get_serial_number': (0, 14, ''),
    'get_sku': (0, 15, ''),

    'set_motor_speed': (1, 4, 'ii'),  # left in mm/s, right in mm/s
    'set_left_motor_speed': (1, 6, 'i'),  # left in mm/s, right in mm/s
    'set_right_motor_speed': (1, 7, 'i'),  # left in mm/s, right in mm/s
    'set_gravity_compensation': (1, 14, 'BH'),  # state (0=off|1=on|2=marker), amount in ‰

    'drive_distance': (1, 8, 'i'),  # distance in mm
    'rotate_angle': (1, 12, 'i'),  # angle in decidegrees
    'drive_arc': (1, 27, 'ii'),  # angle in decidegrees, radius in mm

    'set_marker_eraser': (2, 0, 'B'),  # position (0=up|1=marker down|2=eraser down)
    'get_color_data': (4, 1, '3B'),  # sensor (0 to 3, from left to right), lightning (0=off|1=red|2=green|3=blue|4=all), format (0=12bit|1=mv)

    'set_led_animation': (3, 2, '4B'),  # anim (0=off|1=on|2=blink|3=spin), red, green, blue

    'play_note': (5, 0, 'IH'),  # freq in Hz, duration in ms
    'stop_note': (5, 1, ''),
    'say_phrase': (5, 4, '16s'),  # phrase (bytes, utf8 encoded)

    'get_battery_level': (14, 1, ''),
}

_events = {
    # (dev, cmd): (name, fmt)

    (0, 0): ('version', '9B'),  # board, FWmaj, FWmin, HWmaj, HWmin, Bootmaj, Bootmin, Protomaj, Protomin
    (0, 2): ('name', '16s'),  # name
    (0, 4): ('stopped', ''),
    (0, 11): ('enabled_events', '16s'),  # devices bitfield
    (0, 14): ('serial_number', '12s'),  # serial number
    (0, 15): ('sku', '16s'),  # SKU
    (1, 8): ('drive_distance_finished', ''),
    (1, 12): ('rotate_angle_finished', ''),
    (1, 27): ('drive_arc_finished', ''),
    (1, 29): ('motor_stall', 'I2B'),  # timestamp, motor (0=left|1=right|2=marker), cause
    (2, 0): ('marker_eraser_finished', 'B'),  # position
    (4, 1): ('color_response', '8H'),  # 8 16bit colors
    (4, 2): ('color_event', '16s'),  #
    (5, 0): ('play_note_finished', ''),
    (5, 4): ('say_phrase_finished', ''),
    (12, 0): ('bumper_event', 'IB'),  # timestamp, state(00=no|40=right|80=left|C0=both)
    (13, 0): ('light_event', 'IB2H'),  # timestamp, state, left, right
    (14, 0): ('battery_event', 'IHB'),  # timestamp, voltage, percent
    (14, 1): ('battery_response', 'IHB'),  # timestamp, voltage, percent
    (17, 0): ('touch_event', 'IB'),  # timestamp, state
    (20, 0): ('cliff_event', 'IB2H'),  # timestamp, cliff, sensor, threshold
}


_packet_ids = itertools.count(1)

def format_command(command, *args):
    dev, cmd, fmt = _commands[command]
    pid = next(_packet_ids) % 256
    packet = struct.pack(_msg_prefix + fmt, dev, cmd, pid, *args)

    if len(packet) < SIMPLE_PACKET_LEN:
        packet += bytes(SIMPLE_PACKET_LEN - len(packet))

    packet += crc8(packet).digest()

    assert len(packet) == FULL_PACKET_LEN

    return bytearray(packet), (dev, cmd, pid)


def extract_event(payload):
    dev, cmd, pid = struct.unpack_from(_msg_prefix, payload)
    name, fmt = _events.get((dev, cmd), (f'unknown-{dev}-{cmd}', '16s'))

    if fmt:
        payload = payload[_msg_prefix_len:]
        args = struct.unpack_from(fmt, payload)
    else:
        args = ()

    return name, args, (dev, cmd, pid)
